
#include <stdio.h>
#include <stdlib.h>
#include <string.h>     /* for strlen */

#include <fcntl.h>      /* for open */
#include <unistd.h>     /* for write */

#include <time.h>       /* asctime, gmtime */

#define TRUE    1
#define FALSE   0
#define BOOLEAN int

struct t_mail {
    int id;
    unsigned long size;
    unsigned char *content;
    unsigned long received_time;
    unsigned int colorgroup;
    BOOLEAN is_deleted;
    BOOLEAN was_read;
    BOOLEAN was_answered;
    BOOLEAN is_parked;
    BOOLEAN has_attachement;
    BOOLEAN has_deleted_attachement;
    BOOLEAN is_flagged;
    BOOLEAN was_forwarded;
    
    BOOLEAN high_priority;
    BOOLEAN low_priority;
};

#define IS_DELETED               1
#define IS_READ                  2
#define WAS_ANSWERED             4
#define IS_PARKED                8
#define HAS_ATTACHEMENT         16
#define HAS_DELETED_ATTACHEMENT 32
#define IS_FLAGGED              64
#define WAS_FORWARDED          128


void digit_hexprint(unsigned char character)   {
    printf("%02x ", character);
}

void bitstream_hexprint(unsigned char *buffer, int length)   {
    int i;
    for (i = 0; i < length; i++)  {
        digit_hexprint(buffer[i]);
    }
    printf("\n");
}


BOOLEAN bitstreams_are_equal(unsigned char *buffer1, int length_buffer1, unsigned char *buffer2, int length_buffer2)  {
    if (length_buffer1 != length_buffer2)
        return FALSE;
        
    int i;
    for (i = 0; i < length_buffer1; i++) 
        if (buffer1[i] != buffer2[i])
            return FALSE;
    return TRUE;
}


BOOLEAN is_tbb_file(FILE *filehandle)   {
    int result = TRUE;
    int length_of_magic_number = 6;
    char *tbb_magic_number = (char*) malloc(length_of_magic_number * sizeof(char));
    sprintf(tbb_magic_number, "%c%c%c%c%c%c", 0x20, 0x06, 0x79, 0x19, 0x08, 0x0C);
    unsigned char *first_bytes_from_file = (unsigned char*) malloc(length_of_magic_number * sizeof(char));

    fseek(filehandle, 0, SEEK_SET);
    int read_bytes = fread(first_bytes_from_file, sizeof(char), length_of_magic_number, filehandle);
    if (! bitstreams_are_equal(first_bytes_from_file, read_bytes, tbb_magic_number, length_of_magic_number))
        result = FALSE;
    free(tbb_magic_number);
    free(first_bytes_from_file);
    return result;
}

void skip_global_header(FILE *filehandle)   {
    fseek(filehandle, 3 * 1024 + 2, SEEK_CUR);
}

unsigned char* read_per_mail_header(FILE *filehandle) {
    int sizeof_permail_header = 48; 
    unsigned char *header = malloc(sizeof_permail_header * sizeof(char));
    int read_bytes = fread (header, sizeof(char), sizeof_permail_header, filehandle);
    if (read_bytes == 0 && feof(filehandle))    {
        free(header);
        header = NULL;
        printf("end of file reached!\n");
    }
    else
        bitstream_hexprint(header, read_bytes);
    return header;
}

int is_mail_header(unsigned char *header)    {
    int result = TRUE;
    int length_of_magic_number = 5;
    char *header_magic_number = (char*) malloc(length_of_magic_number * sizeof(char));
    sprintf(header_magic_number, "%c%c%c%c%c", 0x21, 0x9, 0x70, 0x19, 0x30);
    if (! bitstreams_are_equal(header, length_of_magic_number, (unsigned char*) header_magic_number, length_of_magic_number))
        result = FALSE;
    free(header_magic_number);
    return result;
    
}

void init_header(struct t_mail *header) {
    header->id = 0;
    header->size = 0;
    header->received_time = 0;
    header->content = NULL;
    header->is_deleted = FALSE;
    header->was_read = FALSE;
    header->was_answered = FALSE;
    header->is_parked = FALSE;
    header->has_attachement = FALSE;
    header->has_deleted_attachement = FALSE;
    header->is_flagged = FALSE;
    header->was_forwarded = FALSE;
    header->high_priority = FALSE;
    header->low_priority = FALSE;
}

void parse_message_status_flag(struct t_mail *header, char flag) {
    if (flag & IS_DELETED)
        header->is_deleted = TRUE;
    if (flag & IS_READ)
        header->was_read = TRUE;
    if (flag & WAS_ANSWERED)
        header->was_answered = TRUE;
    if (flag & IS_PARKED)
        header->is_parked = TRUE;
    if (flag & HAS_ATTACHEMENT)
        header->has_attachement = TRUE;
    if (flag & HAS_DELETED_ATTACHEMENT)
        header->has_deleted_attachement = TRUE;
    if (flag & IS_FLAGGED)
        header->is_flagged = TRUE;
    if (flag & WAS_FORWARDED)
        header->was_forwarded = TRUE;
}

unsigned long convert_bitstream_to_number(unsigned char *bitstream) {
    int i;
    unsigned long result = 0;
    unsigned long factor = 1;
    for (i = 0; i < 4; i++) {
        result += factor * bitstream[i];
        factor *= 256;
    }
    return result;
}

void parse_mail_priority(struct t_mail *header, unsigned char* bitstream) {
    if (convert_bitstream_to_number(bitstream) == 5) {
        header->high_priority = TRUE;
    }
    if (convert_bitstream_to_number(bitstream) == (((unsigned long) 1) << 32) - 5) {
        header->low_priority = TRUE;
    }
}

struct t_mail* parse_mail_header(unsigned char *headerstring)    {
    struct t_mail *header = malloc(sizeof(struct t_mail));
    if (! is_mail_header(headerstring))   {
        printf("Your TBB file has a currently unknown mail header. Please contact the authors.\n");
        free(header);
        return NULL;
    }
    init_header(header);
    header->received_time = convert_bitstream_to_number(headerstring + 12);
    header->id = 256 * headerstring[17] + headerstring[16];
    parse_message_status_flag(header, headerstring[20]);
    header->colorgroup = convert_bitstream_to_number(headerstring + 28);
    parse_mail_priority(header, headerstring + 32);
    header->size = convert_bitstream_to_number(headerstring + 36);
    return header;
} 

void print_mail_infos(struct t_mail *mail)  {
    printf("id is: %d\n", mail->id);
    printf("received: %lu\n", mail->received_time);
    printf("color group: %x\n", mail->colorgroup);
    printf("size is: %lu bytes\n", mail->size);
    if (mail->is_deleted)
        printf("  deleted\n");
    if (mail->was_read)
        printf("  read\n");
    if (mail->was_answered)
        printf("  answered\n");
    if (mail->is_parked)
        printf("  parked\n");
    if (mail->has_attachement)
        printf("  has attachements\n");
    if (mail->has_deleted_attachement)
        printf("  has deleted attachements\n");
    if (mail->is_flagged)
        printf("  flagged\n");
    if (mail->was_forwarded)
        printf("  forwarded\n");    
    if (mail->high_priority)
        printf("  high priority\n");    
    if (mail->low_priority)
        printf("  low priority\n");    
    printf("\n");
}

struct t_mail* extract_mail(FILE *filehandle) {
    unsigned char *header = read_per_mail_header(filehandle);
    struct t_mail *mail = NULL;
    if (header != NULL) {
        mail = parse_mail_header(header);
        if (mail != NULL)   {
            print_mail_infos(mail);
            mail->content = malloc(mail->size);
            int read_bytes = fread(mail->content, sizeof(char), mail->size, filehandle);
            if (read_bytes != mail->size) {
                printf("TBB file seems to be corrupt! Size of mail was supposed to be %lu bytes but the program could only read %d bytes!\n", mail->size, read_bytes);
                free(mail->content);
                free(mail);
                mail = NULL;
            }
        }
        free(header);
    }
    return mail;
}

int create_mbox_file(void)    {
//    int mbox_file = open("messages.mbox", O_WRONLY | O_CREAT | O_EXCL);
    int mbox_file = open("messages.mbox", O_WRONLY | O_TRUNC | O_EXCL);
    return mbox_file;
}

void write_mbox_fromline(int file, struct t_mail *mail)   {
    char *from = "From - ";
    write(file, from, strlen(from));
    time_t received_time = mail->received_time;
    char *timestring = asctime(gmtime(&received_time));
    write(file, timestring, strlen(timestring));
}

void extract_mails(FILE *tbbfile, int mbox_file)  {
    if (mbox_file < 0)  {
        printf("no mbox file handle!\n");
        return;
    }
    struct t_mail *mail = extract_mail(tbbfile);
    while (mail != NULL)    {
        write_mbox_fromline(mbox_file, mail);
        write(mbox_file, mail->content, mail->size);
        free(mail->content);
        free(mail);
        mail = extract_mail(tbbfile);
    }
}

int main(int argc, char** argv) {
    char *filename = "MESSAGES.TBB";
    FILE *tbbfile = fopen(filename, "r");
    if (tbbfile == NULL)    {
        printf("file '%s' could not be opened!\n", filename);
    }
    else {
        if (is_tbb_file(tbbfile))   {
            printf("file is a TBB file\n");
            skip_global_header(tbbfile);
            int mbox_file = create_mbox_file();
            extract_mails(tbbfile, mbox_file);
            close(mbox_file);
        }
        else {
            printf("file '%s' is not a TBB file (wrong magic number)\n", filename);
        }
        fclose(tbbfile);
    }
    return 0;
}
